{-# LANGUAGE LambdaCase, TypeSynonymInstances, FlexibleInstances, MultiParamTypeClasses, OverloadedStrings #-}
{-# LANGUAGE DuplicateRecordFields, ImportQualifiedPost #-}

module Main where

import GHC.Generics    (Generic)
import Data.Aeson      (FromJSON)
import Data.Binary     (Binary)
import Control.Monad   ((>=>))
import Data.Function   ((&))
import Data.Text       (Text, pack)
import Data.Time       (UTCTime, defaultTimeLocale)
import System.Directory qualified as System
import Data.List       (intersperse)
import Data.Maybe      (fromJust)
import Data.Yaml
import System.FilePath
import Lucid
import Control.Applicative ((<|>))
import Data.Aeson.KeyMap           qualified as KeyMap
import Data.Text.Lazy qualified as LT
import Data.Functor
import Text.Pandoc.Options
import Text.Pandoc.Shared     (isHeaderBlock)
import Text.Pandoc.Definition (Pandoc(Pandoc), Block(Header))

import Achille
import Achille.Writable (Writable)
import Achille.Writable qualified as Writable
import Achille.Internal.IO (AchilleIO)
import Achille.Internal.IO qualified as AchilleIO
import Achille.Task.Pandoc

import Template (Template, Context, parseTemplate)
import Template qualified
import Config (config, ropts, wopts, SiteConfig(title))
import Markdown qualified


-- Bibliography info

data Author = Author
  { name  :: Text
  , url   :: Maybe Text
  , orcid :: Maybe Text
  } deriving (Eq, Show, Generic, Binary)

instance FromJSON Author where
  parseJSON o =
    withObject "Author info"
      (\v -> Author <$> v .: "name"
                    <*> v .:? "url"
                    <*> v .:? "orcid") o
    <|> withText "Author name" (\n -> pure $ Author n Nothing Nothing) o

data Venue = Venue
  { name :: Text
  , url :: Text
  } deriving (Eq, Show, Generic, Binary, FromJSON)

data Publication = Publication
  { title   :: Text
  , slug    :: Text
  , authors :: [Author]
  , file    :: Maybe Text
  , doi     :: Maybe Text
  , venue   :: Venue
  } deriving (Eq, Show, Generic, Binary, FromJSON)

parseYaml :: (AchilleIO m, FromJSON a) => FilePath -> Task m a
parseYaml p = do
  res <- decodeEither' <$> readBS p
  case res of
    Left err -> fail (prettyPrintParseException err)
    Right ok -> return ok

render :: Template -> Context -> Text -> LT.Text
render template ctx body =
  KeyMap.insert "body" (String body) ctx
  & Template.render template

-- writing Html to disk (efficiently)
instance AchilleIO m => Writable m (Html ()) where
    write to = Writable.write to . renderBS

data Note = Note
  { date  :: UTCTime
  , title :: Text
  }
  deriving (Eq, Binary, Generic, FromJSON)

main = achille do
  index <- readTemplate "index.html"

  match_ "assets/*" copyFile
  match_ "papers/*" copyFile

  match_ "static/*" $ copyFileAs (makeRelative "static/")

  (nav, index') <- matchFile "index.md" \src -> do
    doc <- readPandocWith def {readerExtensions = pandocExtensions} src

    let Pandoc _ blocks = doc
    let headers = filter isHeaderBlock blocks

    -- parsing rendered md as template
    let template = renderText (Markdown.render doc) & LT.toStrict & parseTemplate & fromJust

    return (LT.toStrict $ renderText $ renderNav headers, template)


  pubs :: [Publication] <- matchFile "publications.yaml" parseYaml

  notes :: [(FilePath, Note)] <- match "notes/*.md" \src -> do
    if ".lagda.md" /= takeExtensions src then do
      (meta@Note{..}, doc) <- readPandocMetadataWith def {readerExtensions = pandocExtensions} src

      path <- LT.toStrict (renderText (Markdown.render doc))
        & write (src -<.> "html")
        . render index (KeyMap.insert "title" (String title) $ KeyMap.insert "nav" (String nav) $ KeyMap.empty)
      return (path, meta)
    else do
      spath <- toAbsolute src
      odir  <- getOutputDir <&> (</> dropExtensions src)
      tmp   <- liftIO System.getTemporaryDirectory <&> (</> "escot")

      liftIO $ System.createDirectoryIfMissing False tmp
      liftIO $ System.createDirectoryIfMissing True odir
      liftIO $ AchilleIO.copyFile spath (tmp </> "index.lagda.md")

      dir <- liftIO System.getCurrentDirectory
      liftIO $ System.setCurrentDirectory tmp

      callCommand $
                "agda --html "
                <> "--html-dir=. "
                <> "--html-highlight=auto "
                <> "index.lagda.md"

      liftIO $ System.setCurrentDirectory dir
      callCommand $ "cp " <> tmp <> "/* " <> odir

      (meta@Note{..}, doc) <- readAbsPandocMetadataWith def {readerExtensions = pandocExtensions} (tmp </> "index.md")
      path <- renderPandoc doc >>= write (dropExtensions src </> "index.html")
        . render index (KeyMap.insert "title" (String title) $ KeyMap.insert "nav" (String nav) $ KeyMap.empty)
      return (takeDirectory path, meta)

  let notesList :: Html () =
        ul_ $ foldMap (\(src, Note{..}) -> li_ $ a_ [ href_ (pack src) ] $ toHtmlRaw title) notes

  watch pubs $ watch notes do
    let ctx = KeyMap.insert "nav" (String nav) $
              KeyMap.insert "notes" (String (LT.toStrict $ renderText notesList)) $
              KeyMap.insert "title" (String "Lucas Escot") $
              KeyMap.insert "publications" (String (LT.toStrict $ renderText $ renderPublications pubs)) KeyMap.empty

    write "index.html" $ render index ctx (LT.toStrict $ Template.render index' ctx)

  where
    renderNav :: [Block] -> Html ()
    renderNav = ul_ . foldMap (\(Header _ (id, _, _) _) -> li_ $ a_ [href_ $ "/#" <> id] (toHtmlRaw id))

    readTemplate :: FilePath -> Task IO Template
    readTemplate src = readText src <&> parseTemplate >>= \case
      Just t  -> return t
      Nothing -> fail $ "Could not parse template: " ++ src

-- TODO: fix year
renderPublications :: [Publication] -> Html ()
renderPublications pubs = section_ [class_ "pubs"] do
  ul_ [ data_ "year" "2022" ] $ foldMap renderPub pubs

  where renderPub :: Publication -> Html ()
        renderPub Publication {..} = li_ [id_ slug] do
          p_ [class_ "title"] $ a_ [href_ ("#" <> slug)] $ toHtmlRaw title
          p_ [class_ "authors"] $ renderAuthors (reverse (fmap renderAuthor authors))
          let Venue vName vUrl = venue
          p_ [class_ "venue"] $ a_ [href_ vUrl] $ toHtmlRaw vName
          p_ [class_ "buttons"] do
            maybe mempty (\url -> a_ [href_ ("/" <> url)] "PDF") file
            maybe mempty (\doi -> a_ [href_ ("https://doi.org/" <> doi)] "DOI") doi

        renderAuthor :: Author -> Html ()
        renderAuthor (Author name url orcid) =
          case url of
            Nothing  -> toHtmlRaw name
            Just url -> a_ [href_ url] (toHtmlRaw name)

        renderAuthors :: [Html ()] -> Html ()
        renderAuthors [] = mempty
        renderAuthors [one] = one
        renderAuthors (last:others) =
          mconcat (intersperse ", " (reverse others)) <> " and " <> last
